/*
 * MIDIParser.cpp - MIDI Parser File
 *
 * Standard Type 0 MIDI File Parser Code
 *
 * Written by: Craig A. Lindley
 * Version: 1.0
 * Last Update: 09/10/2014
 *
 * Copyright (c) 2014 Craig A. Lindley
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
 * the Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include "MIDIParser.h"
#include "SAM2195.h"
#include "SdFat.h"
#include "SdFatUtil.h"
#include "Utils.h"

#define SMF_HEADER_TAG    "MThd"
#define SMF_HEADER_TAG_LENGTH 4
#define SMF_TRACK_TAG     "MTrk"
#define SMF_TRACK_TAG_LENGTH  4

extern SdFile file;

// States of the MIDI parser finite state machine
enum MP_STATES {
    STATE_INIT, STATE_OPEN_SMF, STATE_PARSE_HEADER, STATE_PARSE_TRACK, STATE_PARSE_EVENT,
    STATE_PARSE_MIDI_EVENT, STATE_PARSE_META_EVENT, STATE_SUCCESS, STATE_ERROR 
};

timeDivisionCallbackType      timeDivisionCallback      = NULL;

// MIDI event callback variables
noteOffCallbackType           noteOffCallback           = NULL;
noteOnCallbackType            noteOnCallback            = NULL;
noteAfterTouchCallbackType    noteAfterTouchCallback    = NULL;
controllerCallbackType        controllerCallback        = NULL;
programChangeCallbackType     programChangeCallback     = NULL;
channelAfterTouchCallbackType channelAfterTouchCallback = NULL;
pitchBendCallbackType         pitchBendCallback         = NULL;

// META event callback variables
sequenceNumberCallbackType    sequenceNumberCallback    = NULL;
textEventCallbackType         textEventCallback         = NULL;
copyrightNoticeCallbackType   copyrightNoticeCallback   = NULL;
sequenceTrackNameCallbackType sequenceTrackNameCallback = NULL;
instrumentNameCallbackType    instrumentNameCallback    = NULL;
lyricCallbackType             lyricCallback             = NULL;
markerCallbackType            markerCallback            = NULL;
cuePointCallbackType          cuePointCallback          = NULL;
midiChannelPrefixCallbackType midiChannelPrefixCallback = NULL;
setTempoCallbackType           setTempoCallback         = NULL;
smpteOffsetCallbackType        smpteOffsetCallback      = NULL;
timeSignatureCallbackType      timeSignatureCallback    = NULL;
keySignatureCallbackType       keySignatureCallback     = NULL;
sequencerSpecificMetaEventCallbackType sequencerSpecificMetaEventCallback = NULL;

void setHandleTimeDivisionCallback(timeDivisionCallbackType func) {    
    timeDivisionCallback = func;
}

// Functions to set MIDI event callbacks
void setHandleNoteOffCallback(noteOffCallbackType func) {    
    noteOffCallback = func;
}

void setHandleNoteOnCallback(noteOnCallbackType func) {    
    noteOnCallback = func;
}

void setHandleNoteAfterTouchCallback(noteAfterTouchCallbackType func) {    
    noteAfterTouchCallback = func;
}

void setHandleControllerCallback(controllerCallbackType func) {    
    controllerCallback = func;
}

void setHandleProgramChangeCallback(programChangeCallbackType func) {    
    programChangeCallback = func;
}

void setHandleChannelAfterTouchCallback(channelAfterTouchCallbackType func) {    
    channelAfterTouchCallback = func;
}

void setHandlePitchBendCallback(pitchBendCallbackType func) {    
    pitchBendCallback = func;
}

// Functions to set META event callbacks
void setHandleSequenceNumberCallback(sequenceNumberCallbackType func) {    
    sequenceNumberCallback = func;
}

void setHandleTextEventCallback(textEventCallbackType func) {    
    textEventCallback = func;
}

void setHandleCopyrightNoticeCallback(copyrightNoticeCallbackType func) {    
    copyrightNoticeCallback = func;
}

void setHandleSequenceTrackNameCallback(sequenceTrackNameCallbackType func) {    
    sequenceTrackNameCallback = func;
}

void setHandleInstrumentNameCallback(instrumentNameCallbackType func) {    
    instrumentNameCallback = func;
}

void setHandleLyricCallback(lyricCallbackType func) {    
    lyricCallback = func;
}

void setHandleMarkerCallback(markerCallbackType func) {    
    markerCallback = func;
}

void setHandleCuePointCallback(cuePointCallbackType func) {    
    cuePointCallback = func;
}

void setHandleMidiChannelPrefixCallback(midiChannelPrefixCallbackType func) {    
    midiChannelPrefixCallback = func;
}

void setHandleSetTempoCallback(setTempoCallbackType func) {    
    setTempoCallback = func;
}

void setHandleSMPTEOffsetCallback(smpteOffsetCallbackType func) {    
    smpteOffsetCallback = func;
}

void setHandleTimeSignatureCallback(timeSignatureCallbackType func) {    
    timeSignatureCallback = func;
}

void setHandleKeySignatureCallback(keySignatureCallbackType func) {    
    keySignatureCallback = func;
}

void setHandleSequencerSpecificMetaEventCallback(sequencerSpecificMetaEventCallbackType func) {    
    sequencerSpecificMetaEventCallback = func;
}

// Peek at the next byte in the file
int peekByte() {

    return file.peek();
}

// Read the next byte in the file
int readByte() {

    int b = file.read();
    if (b == -1) {
        logger(LOG_ERROR, "Read error or EOF occurred");
    }
    return b;
}

// Read a word from the file
int readWord() {

    // SMF is big-endian format
    int b1 = readByte();
    int b0 = readByte();    

    return (b1 << 8) | b0;
}

// Read a double word from the file
int readDWord() {

    // SMF is big-endian format
    int b3 = readByte();
    int b2 = readByte();
    int b1 = readByte();
    int b0 = readByte();    

    return (b3 << 24) | (b2 << 16) | (b1 << 8) | b0;
}

// Read the specified number of bytes from the file into the specified buffer
int readIntoBuffer(void *buffer, int numberOfBytes) {

    int result = file.read(buffer, numberOfBytes);
    if (result == -1) {
        logger(LOG_ERROR, "Read error or EOF occurred");
    }
    return result;
}

// Backup the read stream by n bytes
void backUpStream(int n) {
    file.seekCur(-n);
}

// Return a variable length count from MIDI stream
uint32_t getVariableLengthCount() {

    int b;
    uint32_t count = 0;
    do {
        b = readByte();
        count = (count << 7) + (b & 0x7f);
    } 
    while (b & 0x80);
    return count;
}

// Info extracted from SMF header
int formatType;
int numberOfTracks;
int timeDivision;

// Parser data
int input;
int length;
int trackNumber;
int channel;
int parameter1;
int parameter2;
int parameter3;
int parameter4;
int parameter5;
int temp;
int result;
int tics;
int deltaTime;
int event;
int lastEvent;
int metaCmd;
int metaLength;
char buffer[512];    // Buffer for extracting text strings from meta events

boolean done;

// Terminate the MIDI parser which ends parsing and playback
void terminateMIDIPlayback() {    
    done = true;
}

// Parse and play a SMF using a FSM
int parseSMF (const char *pathname) {

    MP_STATES state = STATE_INIT;
    done = false;

    // Run state machine to completion or error
    while (! done) {
        // Clear text buffer each pass
        memset(buffer, 0, sizeof(buffer));

        switch(state) {

        case STATE_INIT:
            // Close the file in case it was open before
            file.close();

            // Initialize variables
            trackNumber = 0;
            result = ERROR_NONE;
            tics = 0;
            event = lastEvent = 0;

            // Setup next state
            state = STATE_OPEN_SMF;
            break;

        case STATE_OPEN_SMF:
            // Attempt to open the file for reading
            if (! file.open(pathname)) {
                logger(LOG_ERROR, "Error opening SMF file\n");
                result = ERROR_FILEOPEN;
                state = STATE_ERROR;
                continue;
            }    
            logger(LOG_MSG, "SMF opened successfully\n");

            // Adv to next state
            state = STATE_PARSE_HEADER;
            break;

        case STATE_PARSE_HEADER:
            // Read header into memory
            readIntoBuffer(buffer, SMF_HEADER_TAG_LENGTH);

            // Determine if header tag is correct
            if (strncmp(buffer, SMF_HEADER_TAG, SMF_HEADER_TAG_LENGTH) != 0) {
                state = STATE_ERROR;
                result = ERROR_BADFORMAT;
                continue;
            }

            // Determine if header chunk is correct size
            length = readDWord();
            if (length != 6) {
                state = STATE_ERROR;
                result = ERROR_BADFORMAT;
                continue;
            }    
            logger(LOG_MSG, "Got SMF header\n");

            // Read the header content
            formatType = readWord();
            logger(LOG_MSG, "Format type: %d\n", formatType);

            // Check if the SMF is a type 0 (single track) file
            if (formatType != 0) {
                state = STATE_ERROR;
                result = ERROR_FILENOTTYPE0;
                continue;
            }

            numberOfTracks = readWord();
            logger(LOG_MSG, "Number of tracks: %d\n", numberOfTracks);

            // Check for single track
            if (numberOfTracks != 1) {
                state = STATE_ERROR;
                result = ERROR_FILENOTTYPE0;
                continue;
            }

            timeDivision = readWord();
            if (timeDivision & 0x8000) {
                // timeDivision is in frames/second which we cannot handle yet
                state = STATE_ERROR;
                result = ERROR_TIMEDIVISION;
                continue;
            }    
            else    {
                // Report timeDivision in tics/beat or PPQN
                timeDivisionCallback(timeDivision);

                state = STATE_PARSE_TRACK;
            }
            break;

        case STATE_PARSE_TRACK:
            logger(LOG_MSG, "Parsing track number: %d\n", trackNumber++);

            if (trackNumber <= numberOfTracks) {
                // Read header into memory
                readIntoBuffer(buffer, SMF_TRACK_TAG_LENGTH);

                // Determine if track tag is correct
                if (strncmp(buffer, SMF_TRACK_TAG, SMF_TRACK_TAG_LENGTH) != 0) {
                    state = STATE_ERROR;
                    result = ERROR_BADFORMAT;
                    continue;
                }
                logger(LOG_MSG, "Got SMF track\n");                

                length = readDWord();
                logger(LOG_MSG, "Track size: %d\n", length);

                // Setup next state
                state = STATE_PARSE_EVENT;
            }    
            else    {
                state = STATE_SUCCESS;
            }
            break;

        case STATE_PARSE_EVENT:
            logger(LOG_MSG, "Parsing SMF Track Event\n");

            deltaTime = getVariableLengthCount();
            logger(LOG_MSG, "Event delta time: %d\n", deltaTime);
            
            // Check for right joystick activity and terminate if any
            if (pollRightJoystick()) {
                midiReset();
                done = true;
                continue;
            }

            tics += deltaTime;
            logger(LOG_MSG, "tics: %d\n", tics);

            // Check running status
            input = peekByte();
            if (input < 0x80) {
                // Use same event as before
                event = lastEvent;
            }    
            else    {
                // Use new event
                event = readByte();
                lastEvent = event;
            }

            // Determine type of event
            if (event != 0xFF) {
                state = STATE_PARSE_MIDI_EVENT;
            }    
            else    {
                state = STATE_PARSE_META_EVENT;
            }
            break;

        case STATE_PARSE_MIDI_EVENT:
            logger(LOG_MSG, "Processing MIDI event\n");

            // Extract channel from event
            channel = event & 0x0F;

            // Read the first parameter
            parameter1 = readByte();

            switch (event >> 4) {
            case 0x08:    // Note Off
                // Read second parameter
                parameter2 = readByte();

                // Do callback if defined
                if (noteOffCallback != NULL) {
                    noteOffCallback(tics, channel, parameter1, parameter2);
                }
                break;

            case 0x09:    // Note On
                // Read second parameter
                parameter2 = readByte();

                // Do callback if defined
                if (noteOnCallback != NULL) {
                    noteOnCallback(tics, channel, parameter1, parameter2);
                }
                break;

            case 0x0A:    // Note After Touch
                // Read second parameter
                parameter2 = readByte();

                // Do callback if defined
                if (noteAfterTouchCallback != NULL) {
                    noteAfterTouchCallback(tics, channel, parameter1, parameter2);
                }
                break;

            case 0x0B:    // Controller
                // Read second parameter
                parameter2 = readByte();

                // Do callback if defined
                if (controllerCallback != NULL) {
                    controllerCallback(tics, channel, parameter1, parameter2);
                }
                break;

            case 0x0C:    // Program Change
                // Do callback if defined
                if (programChangeCallback != NULL) {
                    programChangeCallback(tics, channel, parameter1);
                }
                break;

            case 0x0D:    // Channel After Touch
                // Do callback if defined
                if (channelAfterTouchCallback != NULL) {
                    channelAfterTouchCallback(tics, channel, parameter1);
                }
                break;

            case 0x0E:    // Pitch Bend
                // Read second parameter
                parameter2 = readByte();
                
                temp = (parameter2 << 7) | (parameter1 & 0x7F);

                // Do callback if defined
                if (pitchBendCallback != NULL) {
                    pitchBendCallback(tics, channel, temp);
                }
                break;

            case 0x0F:
                // Push back the previous parameter1 read so we can read variable length count
                backUpStream(1);

                // Get length of SysEx data
                length = getVariableLengthCount();
                logger(LOG_INFO, "\nlength: %d\n", length);

                // Ignore the data for now
                while (length--) {
                    readByte();
                }
                break;
            }
            // Setup next state
            state = STATE_PARSE_EVENT;
            break;

        case STATE_PARSE_META_EVENT:
            logger(LOG_MSG, "Processing Meta event\n");
            metaCmd = readByte();
            logger(LOG_MSG, "metaCmd: %X\n", metaCmd);
            metaLength = readByte();
            logger(LOG_MSG, "metaLength: %d\n", metaLength);

            switch (metaCmd) {

            case 0x00:    // Sequence number
                parameter1 = readByte();
                parameter2 = readByte();
                // Do callback if defined
                if (sequenceNumberCallback != NULL) {
                    sequenceNumberCallback(tics, parameter1, parameter2);
                }
                break;

            case 0x01:    // Text event
                readIntoBuffer(buffer, metaLength);
                // Do callback if defined
                if (textEventCallback != NULL) {
                    textEventCallback(tics, buffer);
                }
                break;

            case 0x02:    // Copyright Notice
                readIntoBuffer(buffer, metaLength);
                // Do callback if defined
                if (copyrightNoticeCallback != NULL) {
                    copyrightNoticeCallback(tics, buffer);
                }
                break;

            case 0x03:    // Sequence track name
                readIntoBuffer(buffer, metaLength);
                // Do callback if defined
                if (sequenceTrackNameCallback != NULL) {
                    sequenceTrackNameCallback(tics, buffer);
                }
                break;

            case 0x04:    // Instrument name
                readIntoBuffer(buffer, metaLength);
                // Do callback if defined
                if (instrumentNameCallback != NULL) {
                    instrumentNameCallback(tics, buffer);
                }
                break;

            case 0x05:    // Lyric
                readIntoBuffer(buffer, metaLength);
                // Do callback if defined
                if (lyricCallback != NULL) {
                    lyricCallback(tics, buffer);
                }
                break;

            case 0x06:    // Marker
                readIntoBuffer(buffer, metaLength);
                // Do callback if defined
                if (markerCallback != NULL) {
                    markerCallback(tics, buffer);
                }
                break;

            case 0x07:    // Cue point
                readIntoBuffer(buffer, metaLength);
                // Do callback if defined
                if (cuePointCallback != NULL) {
                    cuePointCallback(tics, buffer);
                }
                break;

            case 0x20:    // MIDI channel prefix
                parameter1 = readByte();
                if (midiChannelPrefixCallback != NULL) {
                    midiChannelPrefixCallback(tics, parameter1);   
                }
                break;

            case 0x2F:    // End of Track indication
                logger(LOG_MSG, "End of Track\n");
                // Setup next state
                state = STATE_PARSE_TRACK;
                continue;
                break;

            case 0x51:    // Tempo
                parameter1 = readByte();
                parameter2 = readByte();
                parameter3 = readByte();
                temp = (parameter1 << 16) | (parameter2 << 8) | parameter3;
                // Do callback if defined
                if (setTempoCallback != NULL) {
                    setTempoCallback(tics, temp);
                }
                break;

            case 0x54:    // SMPTE offset
                parameter1 = readByte();
                parameter2 = readByte();
                parameter3 = readByte();
                parameter4 = readByte();
                parameter5 = readByte();
                // Do callback if defined
                if (smpteOffsetCallback != NULL) {
                    smpteOffsetCallback(tics, parameter1, parameter2, parameter3, parameter4, parameter5);
                }
                break;

            case 0x58:    // Time signature
                parameter1 = readByte();
                parameter2 = readByte();
                parameter3 = readByte();
                parameter4 = readByte();
                // Do callback if defined
                if (timeSignatureCallback != NULL) {
                    timeSignatureCallback(tics, parameter1, parameter2, parameter3, parameter4);
                }
                break;

            case 0x59:    // Key signature
                parameter1 = readByte();
                parameter2 = readByte();
                // Do callback if defined
                if (keySignatureCallback != NULL) {
                    keySignatureCallback(tics, parameter1, parameter2);
                }
                break;

            case 0x7F:    // Sequence specific meta event
                readIntoBuffer(buffer, metaLength);
                // Do callback if defined
                if (sequencerSpecificMetaEventCallback != NULL) {
                    sequencerSpecificMetaEventCallback(tics, buffer);
                }
                break;

            default:    // Unknown META event
                readIntoBuffer(buffer, metaLength);
                logger(LOG_WARNING, "Unknown Meta Event: %s\n", buffer);
                break;
            }
            // Setup next state
            state = STATE_PARSE_EVENT;
            break;

        case STATE_SUCCESS:
            logger(LOG_MSG, "Success parsing SMF\n");
            done = true;
            break;

        case STATE_ERROR:
            logger(LOG_ERROR, "Error parsing SMF: %d\n", result);
            done = true;
            break;
        }
    }
    file.close();
    return result;
}



