/*
 * Utils.cpp - Utility Functions
 *
 * Written by: Craig A. Lindley
 * Version: 1.0
 * Last Update: 09/10/2014
 *
 * Copyright (c) 2014 Craig A. Lindley
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
 * the Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include "Arduino.h"
#include "Utils.h"

// Limit to the size of printed string with full variable expansion
#define PRINTF_BUFFER_SIZE 256

char buf[PRINTF_BUFFER_SIZE];

// printf for Arduino
void myprintf(char *fmt, ... ) {

    va_list args;
    va_start (args, fmt);
    vsnprintf(buf, PRINTF_BUFFER_SIZE, fmt, args);
    va_end (args);

    Serial.print(buf);
    Serial.flush();

    // Delay so as not to over run the Serial connection
    delay(20);
}

// Current logging level
int logLevel = LOG_MSG;

// Function for setting log level
void setLogLevel(int _logLevel) {
    logLevel = _logLevel;
}

// Logging function associates a log level with log string
void logger(int level, char *fmt, ...) {

    // Return if level is insufficient
    if (level < logLevel) {
        return;
    }
    va_list args;
    va_start (args, fmt);
    vsnprintf(buf, PRINTF_BUFFER_SIZE, fmt, args);
    va_end (args);

    Serial.print(buf);
    Serial.flush();

    // Delay so as not to over run the Serial connection
    delay(10);
}

boolean pollLeftJoystickButton() {

    return (digitalRead(JOYSTICK_LEFT_BUTTON) == 0);
}

void waitForLeftJoystickButton() {

    while (digitalRead(JOYSTICK_LEFT_BUTTON) == 1) {
        delay(100);
    }
}

boolean pollRightJoystickButton() {

    return (digitalRead(JOYSTICK_RIGHT_BUTTON) == 0);
}

void waitForRightJoystickButton() {

    while (digitalRead(JOYSTICK_RIGHT_BUTTON) == 1) {
        delay(100);
    }
}

// Read joystick pin
// Return 0 if joystick released
// Return 1 if joystick is moved towards lower resistance
// Return 2 if joystick is moved towards higher resistance
int _readJoystick(int inputPin) {

    int value = analogRead(inputPin);
    if (value < 300) {
        return 1;
    }    
    else if (value > 700) {
        return 2;
    }    
    else    {
        return 0;
    }
}

// Poll joystick for activity
// Returns true if joystick is active
boolean _pollJoystick(int STS, int TTB) {
    
    if ((_readJoystick(STS) != 0) || (_readJoystick(TTB) != 0)) {
        return true;
    }    else    {
        return false;
    }
}

// Read the side to side (STS) pot and the top to bottom (TTB) pot
// of the joystick to return the direction the joystick was moved
enum JS_DIRECTION _readJoystick(int STS, int TTB) {

    int input;
    int count = 0;

    while (count == 0) {
        input = _readJoystick(STS);
        if (input != 0) {
            count += input;
        }
        delay(100);
        input = _readJoystick(TTB);
        if (input != 0) {
            count += input + 2;
        }
    }

    // Wait for joystick to be released
    while ((_readJoystick(STS) != 0) || (_readJoystick(TTB) != 0)) {
        delay(10);
    }
    // Convert count to direction
    switch (count) {
        case 4: 
            return DIR_UP;
            
        case 3:
            return DIR_DOWN;
            
        case 2:
            return DIR_RIGHT;
            
        case 1:
            return DIR_LEFT;
    }
}

boolean pollLeftJoystick() {
    return _pollJoystick(JOYSTICK_LEFT_STS, JOYSTICK_LEFT_TTB);
}

enum JS_DIRECTION readLeftJoystick() {
    return _readJoystick(JOYSTICK_LEFT_STS, JOYSTICK_LEFT_TTB);
}

boolean pollRightJoystick() {
    return _pollJoystick(JOYSTICK_RIGHT_STS, JOYSTICK_RIGHT_TTB);
}

enum JS_DIRECTION readRightJoystick() {
    return _readJoystick(JOYSTICK_RIGHT_STS, JOYSTICK_RIGHT_TTB);
}



